import re
import warnings
from typing import Any, Dict, Iterable, List, Optional, Tuple, Type, Union
from uuid import UUID

import requests

from ..util.cheapdantic import BaseModel
from .exc import OdpResourceExistsError, OdpResourceNotFoundError, OdpValidationError
from .http_client import OdpHttpClient
from odp.dto import ResourceDto


class OdpResourceClient(BaseModel):
    """Client for interacting with ODP resources."""

    http_client: OdpHttpClient
    resource_endpoint: str

    def __init__(self, **kwargs):
        if not re.match(r"^\/[/.a-zA-Z0-9-]+$", kwargs.get("resource_endpoint")):
            raise ValueError(f"Invalid resource endpoint: {kwargs.get('resource_endpoint')}")
        super().__init__(**kwargs)

    @property
    def resource_url(self) -> str:
        """The URL of the resource endpoint, including the base URL.

        Returns:
            The resource URL
        """
        return f"{self.http_client.base_url}{self.resource_endpoint}"

    def get(
        self,
        ref: Union[UUID, str],
        tp: Optional[Type[ResourceDto]] = None,
        assert_type: bool = False,
        raise_unknown_kind: bool = False,
    ) -> ResourceDto:
        """Get a resource by reference.

        The reference can be either a UUID or a qualified name.

        Args:
            ref: Resource reference
            tp: Optionally cast the fetched resource to a specific type
            assert_type: Whether to assert that the fetched resource is of the expected type, must be used with `tp`
            raise_unknown_kind: Whether to raise an error if the kind of the resource is not known

        Returns:
            The manifest of the resource corresponding to the reference

        Raises:
            OdpResourceNotFoundError: If the resource does not exist
            OdpValidationError: Invalid input
        """
        warnings.warn(
            "catalog class is deprecated and will be removed in a future versions. Use catalog_v2 class instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        res = self.http_client.get(f"{self.resource_endpoint}/{ref}")

        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input") from e
            if res.status_code == 404:
                raise OdpResourceNotFoundError(f"Resource not found: {ref}") from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")

        return ResourceDto.create(**res.json())

    def list(
        self,
        oqs_filter: Optional[Dict[str, Any]] = None,
        cursor: Optional[str] = None,
        tp: Optional[Type[ResourceDto]] = None,
        assert_type: bool = False,
        raise_unknown_kind: bool = False,
    ) -> Iterable[ResourceDto]:
        """List all resources based on the provided filter

        Args:
            oqs_filter: OQS filter
            cursor: Optional cursor for pagination
            tp: Optionally cast the fetched resource to a specific type
            assert_type: Whether to assert that the fetched resource is of the expected type, must be used with `tp`
            raise_unknown_kind: Whether to raise an error if the kind of the resource is not known

        Yields:
            Resources matching the provided filter
        """
        warnings.warn(
            "catalog class is deprecated and will be removed in a future versions. Use catalog_v2 class instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        while True:
            page, cursor = self.list_paginated(
                oqs_filter=oqs_filter,
                cursor=cursor,
                tp=tp,
                assert_type=assert_type,
                raise_unknown_kind=raise_unknown_kind,
            )

            yield from page
            if not cursor:
                break

    def list_paginated(
        self,
        oqs_filter: Optional[dict] = None,
        cursor: Optional[str] = None,
        limit: int = 1000,
        tp: Optional[Type[ResourceDto]] = None,
        assert_type: bool = False,
        raise_unknown_kind: bool = False,
    ) -> Tuple[List[ResourceDto], Optional[str]]:
        """List a page of resources based on the provided filter

        Args:
            oqs_filter: OQS filter
            cursor: Cursor for pagination
            limit: Maximum number of resources to return
            tp: Optionally cast the fetched resource to a specific type
            assert_type: Whether to assert that the fetched resource is of the expected type, must be used with `tp`
            raise_unknown_kind: Whether to raise an error if the kind of the resource is not known

        Returns:
            A page of resources

        Raises:
            OdpValidationError: Invalid input
        """

        warnings.warn(
            "catalog class is deprecated and will be removed in a future versions. Use catalog_v2 class instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        params = {}
        body = None

        if cursor:
            params["page"] = cursor
        if limit:
            params["page_size"] = limit

        if oqs_filter:
            body = oqs_filter

        res = self.http_client.post(self.resource_endpoint + "/list", params=params, content=body)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 401:
                raise OdpValidationError("API argument error") from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")

        content = res.json()
        return [ResourceDto.create(**item) for item in content["results"]], content.get("next")

    def create(
        self,
        manifest: ResourceDto,
        assert_type: bool = False,
        raise_unknown_kind: bool = False,
    ) -> ResourceDto:
        """Create a resource from a manifest

        Args:
            manifest: Resource manifest
            assert_type: Whether to assert the type of the object returned by the API
            raise_unknown_kind: Whether to raise an error if the kind of the resource is not known

        Returns:
            The manifest of the created resource, populated with uuid and status

        Raises:
            OdpValidationError: Invalid input
            OdpResourceExistsError: Resource already exists
        """

        warnings.warn(
            "catalog class is deprecated and will be removed in a future versions. Use catalog_v2 class instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        res = self.http_client.post(self.resource_endpoint, content=manifest)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input", res.text) from e
            if res.status_code == 409:
                raise OdpResourceExistsError("Resource already exists") from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")

        return ResourceDto.create(**res.json())

    def update(
        self,
        manifest_update: Union[ResourceDto, dict],
        ref: Union[str, UUID, None] = None,
        tp: Optional[Type[ResourceDto]] = None,
        assert_type: bool = False,
        raise_unknown_kind: bool = False,
    ) -> ResourceDto:
        """Update a resource from a manifest

        Args:
            manifest_update: Resource manifest or JSON patch
            ref: Optional reference to the resource to update.
            tp: Optionally cast the fetched resource to a specific type. Can only be used if `ref` is a UUID.
            assert_type: Whether to assert the type of the object returned by the API
            raise_unknown_kind: Whether to raise an error if the kind of the resource is not known

        Returns:
            The manifest of the updated resource, populated with the updated fields

        Raises:
            OdpValidationError: Invalid input
            OdpResourceNotFoundError: Resource not found
        """
        # if isinstance(manifest_update, ResourceDto) and tp:
        #     raise ValueError("Cannot cast the updated resource to a specific type if the manifest is a ResourceDto")
        # elif isinstance(manifest_update, ResourceDto):
        #     tp = ResourceDto[get_resource_spec_type(manifest_update)]

        warnings.warn(
            "catalog class is deprecated and will be removed in a future versions. Use catalog_v2 class instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        if ref:
            if isinstance(ref, UUID):
                params = {"either_id": str(ref)}
            else:
                rg, rt, name = ref.split("/")
                params = {"either_id": name, "kind": f"{rg}/{rt}"}
        else:
            params = {}

        res = self.http_client.patch(self.resource_endpoint, params=params, content=manifest_update)
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 400:
                raise OdpValidationError("Invalid input") from e
            if res.status_code == 404:
                raise OdpResourceNotFoundError("Resource not found") from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")

        return ResourceDto.create(**res.json())

    def delete(self, ref: Union[UUID, str, ResourceDto]):
        """Delete a resource by reference.

        Args:
            ref: Resource reference. If a `ResourceDto` is passed, the reference will be extracted from the metadata.

        Raises:
            OdpResourceNotFoundError: If the resource does not exist
        """

        warnings.warn(
            "catalog class is deprecated and will be removed in a future versions. Use catalog_v2 class instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )

        if isinstance(ref, ResourceDto):
            ref = ref.metadata.uuid or f"{ref.kind}/{ref.metadata.name}"

        res = self.http_client.delete(f"{self.resource_endpoint}/{ref}")
        try:
            res.raise_for_status()
        except requests.HTTPError as e:
            if res.status_code == 404:
                raise OdpResourceNotFoundError(f"Resource not found: {ref}") from e
            raise requests.HTTPError(f"HTTP Error - {res.status_code}: {res.text}")



from odp.dto import MetadataDto
from odp.dto.registry import ObservableClassDto, ObservableClassSpec

static_single_value_class = ObservableClassDto(
    metadata=MetadataDto(
        name="static-observable",
        labels={"catalog.hubocean.io/released": True},
    ),
    spec=ObservableClassSpec(
        observable_schema={
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "StaticObservable",
            "type": "object",
            "description": "Single value observable",
            "properties": {"attribute": {"title": "Attribute", "type": "string"}, "value": {"title": "Value"}},
            "required": ["value"],
        }
    ),
)


static_coverage_class = ObservableClassDto(
    metadata=MetadataDto(
        name="static-coverage",
        labels={"catalog.hubocean.io/released": True},
    ),
    spec=ObservableClassSpec(
        observable_schema={
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "title": "StaticCoverage",
            "description": "1D real coverage",
            "required": ["attribute", "value"],
            "properties": {
                "value": {
                    "type": "array",
                    "items": [{"type": "number"}, {"type": "number"}],
                    "title": "Value",
                    "maxItems": 2,
                    "minItems": 2,
                },
                "attribute": {"type": "string", "title": "Attribute"},
            },
        }
    ),
)

static_geometric_coverage_class = ObservableClassDto(
    metadata=MetadataDto(
        name="static-geometric-coverage",
        labels={"catalog.hubocean.io/released": True},
    ),
    spec=ObservableClassSpec(
        observable_schema={
            "$schema": "http://json-schema.org/draft-04/schema#",
            "title": "StaticGeometricCoverage",
            "description": "Geometric coverage",
            "type": "object",
            "required": ["attribute", "value"],
            "properties": {
                "attribute": {"title": "Attribute", "type": "string"},
                "value": {"$ref": "#/definitions/Feature"},
            },
            "definitions": {
                "Point": {
                    "title": "Point",
                    "description": "Point Model",
                    "type": "object",
                    "properties": {
                        "type": {"title": "Type", "default": "Point", "const": "Point", "type": "string"},
                        "coordinates": {
                            "title": "Coordinates",
                            "anyOf": [
                                {
                                    "type": "array",
                                    "minItems": 2,
                                    "maxItems": 2,
                                    "items": [{"type": "number"}, {"type": "number"}],
                                },
                                {
                                    "type": "array",
                                    "minItems": 3,
                                    "maxItems": 3,
                                    "items": [{"type": "number"}, {"type": "number"}, {"type": "number"}],
                                },
                            ],
                        },
                    },
                    "required": ["coordinates"],
                },
                "MultiPoint": {
                    "title": "MultiPoint",
                    "description": "MultiPoint Model",
                    "type": "object",
                    "properties": {
                        "type": {"title": "Type", "default": "MultiPoint", "const": "MultiPoint", "type": "string"},
                        "coordinates": {
                            "title": "Coordinates",
                            "minItems": 1,
                            "type": "array",
                            "items": {
                                "anyOf": [
                                    {
                                        "type": "array",
                                        "minItems": 2,
                                        "maxItems": 2,
                                        "items": [{"type": "number"}, {"type": "number"}],
                                    },
                                    {
                                        "type": "array",
                                        "minItems": 3,
                                        "maxItems": 3,
                                        "items": [{"type": "number"}, {"type": "number"}, {"type": "number"}],
                                    },
                                ]
                            },
                        },
                    },
                    "required": ["coordinates"],
                },
                "LineString": {
                    "title": "LineString",
                    "description": "LineString Model",
                    "type": "object",
                    "properties": {
                        "type": {"title": "Type", "default": "LineString", "const": "LineString", "type": "string"},
                        "coordinates": {
                            "title": "Coordinates",
                            "minItems": 2,
                            "type": "array",
                            "items": {
                                "anyOf": [
                                    {
                                        "type": "array",
                                        "minItems": 2,
                                        "maxItems": 2,
                                        "items": [{"type": "number"}, {"type": "number"}],
                                    },
                                    {
                                        "type": "array",
                                        "minItems": 3,
                                        "maxItems": 3,
                                        "items": [{"type": "number"}, {"type": "number"}, {"type": "number"}],
                                    },
                                ]
                            },
                        },
                    },
                    "required": ["coordinates"],
                },
                "MultiLineString": {
                    "title": "MultiLineString",
                    "description": "MultiLineString Model",
                    "type": "object",
                    "properties": {
                        "type": {
                            "title": "Type",
                            "default": "MultiLineString",
                            "const": "MultiLineString",
                            "type": "string",
                        },
                        "coordinates": {
                            "title": "Coordinates",
                            "minItems": 1,
                            "type": "array",
                            "items": {
                                "type": "array",
                                "items": {
                                    "anyOf": [
                                        {
                                            "type": "array",
                                            "minItems": 2,
                                            "maxItems": 2,
                                            "items": [{"type": "number"}, {"type": "number"}],
                                        },
                                        {
                                            "type": "array",
                                            "minItems": 3,
                                            "maxItems": 3,
                                            "items": [{"type": "number"}, {"type": "number"}, {"type": "number"}],
                                        },
                                    ]
                                },
                                "minItems": 2,
                            },
                        },
                    },
                    "required": ["coordinates"],
                },
                "Polygon": {
                    "title": "Polygon",
                    "description": "Polygon Model",
                    "type": "object",
                    "properties": {
                        "type": {"title": "Type", "default": "Polygon", "const": "Polygon", "type": "string"},
                        "coordinates": {
                            "title": "Coordinates",
                            "minItems": 1,
                            "type": "array",
                            "items": {
                                "type": "array",
                                "items": {
                                    "anyOf": [
                                        {
                                            "type": "array",
                                            "minItems": 2,
                                            "maxItems": 2,
                                            "items": [{"type": "number"}, {"type": "number"}],
                                        },
                                        {
                                            "type": "array",
                                            "minItems": 3,
                                            "maxItems": 3,
                                            "items": [{"type": "number"}, {"type": "number"}, {"type": "number"}],
                                        },
                                    ]
                                },
                                "minItems": 4,
                            },
                        },
                    },
                    "required": ["coordinates"],
                },
                "MultiPolygon": {
                    "title": "MultiPolygon",
                    "description": "MultiPolygon Model",
                    "type": "object",
                    "properties": {
                        "type": {"title": "Type", "default": "MultiPolygon", "const": "MultiPolygon", "type": "string"},
                        "coordinates": {
                            "title": "Coordinates",
                            "minItems": 1,
                            "type": "array",
                            "items": {
                                "type": "array",
                                "items": {
                                    "type": "array",
                                    "items": {
                                        "anyOf": [
                                            {
                                                "type": "array",
                                                "minItems": 2,
                                                "maxItems": 2,
                                                "items": [{"type": "number"}, {"type": "number"}],
                                            },
                                            {
                                                "type": "array",
                                                "minItems": 3,
                                                "maxItems": 3,
                                                "items": [{"type": "number"}, {"type": "number"}, {"type": "number"}],
                                            },
                                        ]
                                    },
                                    "minItems": 4,
                                },
                                "minItems": 1,
                            },
                        },
                    },
                    "required": ["coordinates"],
                },
                "GeometryCollection": {
                    "title": "GeometryCollection",
                    "description": "GeometryCollection Model",
                    "type": "object",
                    "properties": {
                        "type": {
                            "title": "Type",
                            "default": "GeometryCollection",
                            "const": "GeometryCollection",
                            "type": "string",
                        },
                        "geometries": {
                            "title": "Geometries",
                            "type": "array",
                            "items": {
                                "anyOf": [
                                    {"$ref": "#/definitions/Point"},
                                    {"$ref": "#/definitions/MultiPoint"},
                                    {"$ref": "#/definitions/LineString"},
                                    {"$ref": "#/definitions/MultiLineString"},
                                    {"$ref": "#/definitions/Polygon"},
                                    {"$ref": "#/definitions/MultiPolygon"},
                                ]
                            },
                        },
                    },
                    "required": ["geometries"],
                },
                "BaseModel": {"title": "BaseModel", "type": "object", "properties": {}},
                "Feature": {
                    "title": "Feature",
                    "description": "Feature Model",
                    "type": "object",
                    "properties": {
                        "type": {"title": "Type", "default": "Feature", "const": "Feature", "type": "string"},
                        "geometry": {
                            "title": "Geometry",
                            "anyOf": [
                                {"$ref": "#/definitions/Point"},
                                {"$ref": "#/definitions/MultiPoint"},
                                {"$ref": "#/definitions/LineString"},
                                {"$ref": "#/definitions/MultiLineString"},
                                {"$ref": "#/definitions/Polygon"},
                                {"$ref": "#/definitions/MultiPolygon"},
                                {"$ref": "#/definitions/GeometryCollection"},
                            ],
                        },
                        "properties": {
                            "title": "Properties",
                            "anyOf": [{"type": "object"}, {"$ref": "#/definitions/BaseModel"}],
                        },
                        "id": {"title": "Id", "type": "string"},
                        "bbox": {
                            "title": "Bbox",
                            "anyOf": [
                                {
                                    "type": "array",
                                    "minItems": 4,
                                    "maxItems": 4,
                                    "items": [
                                        {"type": "number"},
                                        {"type": "number"},
                                        {"type": "number"},
                                        {"type": "number"},
                                    ],
                                },
                                {
                                    "type": "array",
                                    "minItems": 6,
                                    "maxItems": 6,
                                    "items": [
                                        {"type": "number"},
                                        {"type": "number"},
                                        {"type": "number"},
                                        {"type": "number"},
                                        {"type": "number"},
                                        {"type": "number"},
                                    ],
                                },
                            ],
                        },
                    },
                },
            },
        }
    ),
)



from typing import Optional, Union
from uuid import UUID

from ..tabular_v2.client import Table
from ..util.cheapdantic import BaseModel, Field, PrivateAttr
from .auth import ApiKeyTokenProvider, TokenProvider, get_default_token_provider
from .http_client import OdpHttpClient
from .iam_client import OdpIamClient
from .raw_storage_client import OdpRawStorageClient
from .resource_client import OdpResourceClient
from .tabular_storage_client import OdpTabularStorageClient
from .tabular_storage_v2_client import ClientAuthorization
from odp.client.catalog_client import CatalogClient
from odp.dto import DatasetDto


class OdpClient(BaseModel):
    """Client for the ODP API"""

    base_url: str = "https://api.hubocean.earth"
    token_provider: TokenProvider = Field(default_factory=get_default_token_provider)
    api_key: Optional[str] = None

    _http_client: OdpHttpClient = PrivateAttr(None)
    _catalog_client: OdpResourceClient = PrivateAttr(None)
    _raw_storage_client: OdpRawStorageClient = PrivateAttr(None)
    _tabular_storage_client: OdpTabularStorageClient = PrivateAttr(None)
    _tabular_storage_v2_client: ClientAuthorization = PrivateAttr(None)
    _iam_client: OdpIamClient = PrivateAttr(None)
    _catalog_v2_client: CatalogClient = PrivateAttr(None)

    def __init__(self, **data):
        super().__init__(**data)

        if self.api_key:
            self.token_provider = ApiKeyTokenProvider(self.base_url, self.api_key)
        self._http_client = OdpHttpClient(base_url=self.base_url, token_provider=self.token_provider)
        self._catalog_client = OdpResourceClient(http_client=self._http_client, resource_endpoint="/catalog")
        self._raw_storage_client = OdpRawStorageClient(http_client=self._http_client)
        self._tabular_storage_client = OdpTabularStorageClient(http_client=self._http_client)
        self._tabular_storage_v2_client = ClientAuthorization(
            base_url=self.base_url, token_provider=self.token_provider
        )
        self._iam_client = OdpIamClient(http_client=self._http_client)
        self._catalog_v2_client = CatalogClient(http_client=self._http_client)

    def personalize_name(self, name: str, fmt: Optional[str] = None) -> str:
        """Personalize a name by adding a postfix unique to the user

        Args:
            name: The name to personalize
            fmt: Used to override the default format string. Should be a python format-string with placeholders
                for the variables `uid` and `name`. For example: `"{uid}-{name}"`

        Returns:
            The personalized name
        """
        fmt = fmt or "{name}-{uid}"
        uid = self.token_provider.get_user_id()

        # Attempt to simplify the UID by only using the node part of the UUID
        try:
            uid = UUID(uid).node
        except ValueError:
            # User ID is not a valid UUID, use it as-is
            pass

        return fmt.format(uid=uid, name=name)

    @property
    def resource_store(self):
        # TODO: Implement resource store
        raise NotImplementedError("Resource store not implemented")

    @property
    def catalog(self) -> OdpResourceClient:
        return self._catalog_client

    @property
    def catalog_v2_experimental(self) -> CatalogClient:
        """Experimental catalog v2 client"""
        return self._catalog_v2_client

    @property
    def catalog_v2(self) -> CatalogClient:
        """Catalog v2 client"""
        return self._catalog_v2_client

    @property
    def iam(self) -> OdpIamClient:
        return self._iam_client

    @property
    def registry(self):
        # TODO: Implement registry/core controller
        raise NotImplementedError("Registry not implemented")

    @property
    def raw(self) -> OdpRawStorageClient:
        return self._raw_storage_client

    @property
    def tabular(self) -> OdpTabularStorageClient:
        return self._tabular_storage_client

    def table_v2(self, dataset_dto: Union[DatasetDto, str]) -> Table:
        if isinstance(dataset_dto, str):
            return self._tabular_storage_v2_client.table(dataset_dto)
        return self._tabular_storage_v2_client.table(str(dataset_dto.uuid))



import logging
from typing import Iterator


class Reader:
    def read(self, size: int = -1) -> bytes:
        raise NotImplementedError()


class Writer:
    def write(self, data: bytes):
        raise NotImplementedError()

    def close(self):
        pass


class Iter2Reader(Reader):
    """
    convert a byte iterator to a file-like object
    reads will attempt to read the next bytes from the iterator when needed
    FIXME: seems broken when used with real cases, avoid using
    """

    def __init__(self, i: Iterator[bytes]):
        self.iter = i
        self.closed = False
        self.buf = b""

    def preload(self):
        if not self.iter:
            return self
        try:
            self.buf += next(self.iter)
        except StopIteration:
            self.iter = None
        return self

    def read_some(self) -> bytes:
        if not self.buf:
            self.preload()
        out = self.buf
        self.buf = b""
        logging.debug("read %d", len(out))
        return out

    def read(self, size: int = -1) -> bytes:
        # logging.debug("reading...")
        if size < 0:
            return self.read_some()
        while len(self.buf) < size:
            if not self.iter:
                break
            self.preload()
        ret = self.buf[:size]
        self.buf = self.buf[len(ret) :]
        # logging.debug("read %d", len(ret))
        return ret



import pyarrow as pa


class WriteBuffer:
    """used internally to capture the output of a stream in a memory efficient way"""

    def __init__(self):
        self.queue = []
        self.buf = b""
        self.closed = False

    def write(self, data: bytes | pa.Buffer):
        if isinstance(data, pa.Buffer):
            x: pa.Buffer = data
            data = x.to_pybytes()

        if len(self.buf) + len(data) > 1024 * 1024:
            self.queue.append(self.buf)
            self.buf = b""
        # concatenate small chunks
        self.buf += data

    def writable(self) -> bool:
        return True

    def readable(self) -> bool:
        return False

    def flush(self, partial: bool = False) -> list[bytes]:
        out = self.queue
        self.queue = []
        if not partial and self.buf != b"":
            out.append(self.buf)
            self.buf = b""
        return out



"""Authentication handling in the form of token-providers"""
import base64
import json
import logging
import os
import re
import sys
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional, Union
from uuid import UUID

import jwt
import msal
import msal_extensions
import requests
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicNumbers
from requests.auth import AuthBase

from ..util.cheapdantic import BaseModel, PrivateAttr, SecretStr
from .exc import OdpAuthError
from .utils import get_version

LOG = logging.getLogger(__name__)


class TokenProvider(AuthBase, BaseModel, ABC):
    """Base class for token providers"""

    user_agent: str = "odp-sdk/" + get_version()

    @abstractmethod
    def get_token(self) -> str:
        """Returns the token to be used for authentication

        Returns:
            str: The token to be used for authentication

        Raises:
            OdpAuthError: If the token cannot be retrieved
        """

    @abstractmethod
    def get_user_id(self) -> str:
        """Returns the user id to be used for authentication

        Returns:
            str: The user id to be used for authentication

        Raises:
            OdpAuthError: If user currently not authenticated
        """

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        r.headers["Authorization"] = self.get_token()
        return r


class OdpWorkspaceTokenProvider(TokenProvider):
    """Token provider for ODP workspaces"""

    token_uri: str = "http://localhost:8000/access_token"
    """Token endpoint."""

    user_id_claim: str = "sub"
    """The claim to use as the user ID"""

    _user_id: Optional[str] = PrivateAttr(None)

    def __init__(self, **data):
        super().__init__(**data)
        self.user_agent = self.user_agent + " (Workspaces)"

    def get_token(self) -> str:
        res = requests.post(self.token_uri)
        res.raise_for_status()

        token: str = res.json()["token"]
        claims = jwt.decode(token.replace("Bearer ", ""), options={"verify_signature": False})
        self._user_id = claims[self.user_id_claim]

        return "Bearer " + res.json()["token"]

    def get_user_id(self) -> str:
        if self._user_id is None:
            self.get_token()  # This will set the user_id
        return self._user_id


class HardcodedTokenProvider(TokenProvider):
    """Token provider for hardcoded tokens"""

    _token: str = PrivateAttr(None)
    _user_id: Optional[str]

    def __init__(self, token: str, **data):
        super().__init__(**data)
        self._token = token
        self.user_agent = self.user_agent + " (Hardcoded)"

    def get_token(self) -> str:
        return self._token

    def get_user_id(self) -> str:
        if self._user_id is None:
            self._user_id = self._obtain_user_id()

        return self._user_id

    def _obtain_user_id(self) -> str:
        token = self.get_token()
        if token.startswith("Bearer "):
            token = token[7:]
        try:
            claims = jwt.decode(token, options={"verify_signature": False})
        except jwt.exceptions.PyJWTError as e:
            raise OdpAuthError("Unable to obtain user ID from token") from e

        return claims["sub"]


class ApiKeyTokenProvider(TokenProvider):
    """Token provider for API keys"""

    _base_url: str = PrivateAttr(None)
    _api_key: str = PrivateAttr(None)
    _user_id: Optional[str]

    def __init__(self, base_url: str, api_key: str, **data):
        super().__init__(**data)
        self._base_url = base_url
        self._api_key = api_key
        self.user_agent = self.user_agent + " (API Key)"

    def get_token(self) -> str:
        return "ApiKey " + self._api_key

    def get_user_id(self) -> str:
        if self._user_id is None:
            self._user_id = self._get_user_id()
        return self._user_id

    def _get_user_id(self) -> str:
        if not self._api_key:
            raise OdpAuthError("Missing API key")
        return str(UUID(self._api_key))


class JwtTokenProvider(TokenProvider, ABC):
    """Token provider for JWT tokens"""

    audience: Optional[str] = None
    """IDP token audience

    Must be set if the token is to be validated
    """

    validate_token: bool = False
    """Whether to validate the token or not"""

    token_exp_lee_way: int = 300
    """Number of seconds before token expiry we should refresh the token"""

    user_id_claim: str = "sub"
    """The claim to use as the user ID"""

    _access_token: str = PrivateAttr(None)
    _refresh_token: Optional[str] = PrivateAttr(None)
    _claims: Dict[str, Union[int, str]] = PrivateAttr(None)
    _jwks: Dict[str, str] = PrivateAttr(None)
    _expiry: int = PrivateAttr(0)
    _user_id: Optional[str] = PrivateAttr(None)

    @abstractmethod
    def authenticate(self) -> Dict[str, str]:
        """Authenticate against the IDP and return the token

        Returns:
            The token response from the IDP

        Raises:
            OdpAuthError: If the token cannot be retrieved
        """

    @abstractmethod
    def get_jwks_uri(self) -> str:
        """Get the jwks uri for the identity provider

        Returns:
            The jwks uri
        """

    def get_token(self) -> str:
        """Returns the token to be used for authentication

        Returns:
            The token to be used for authentication

        Raises:
            OdpAuthError: If the token cannot be retrieved
        """

        if self._access_token and time.time() < self._expiry - self.token_exp_lee_way:
            access_token = self._access_token
        else:
            auth_response = self.authenticate()
            access_token = self._parse_token(auth_response)
            self._user_id = self._claims[self.user_id_claim]

        return "Bearer {}".format(access_token)

    def get_user_id(self) -> str:
        if not self._user_id:
            self.get_token()  # This will set the user_id
            if not self._user_id:
                raise OdpAuthError("Unable to obtain user ID")

        return self._user_id

    def _parse_and_validate(self, access_token: str) -> str:
        token_components = [self._base64_decode(x) for x in access_token.split(".")]
        headers = json.loads(token_components[0])

        if not headers:
            raise OdpAuthError("No headers in token")

        kid = headers.pop("kid")
        jwk = self._get_jwk(kid)

        pem = (
            RSAPublicNumbers(n=self._decode_jwk_value(jwk["n"]), e=self._decode_jwk_value(jwk["e"]))
            .public_key(default_backend())
            .public_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PublicFormat.SubjectPublicKeyInfo,
            )
        )

        self._claims = jwt.decode(
            access_token,
            pem,
            verify=True,
            options=headers,
            audience=self.audience,
            algorithms=["RS256"],
        )

        self._access_token = access_token
        self._expiry = self._claims["exp"]

        return access_token

    def _parse_novalidate(self, access_token: str) -> str:
        self._claims = jwt.decode(access_token, options={"verify_signature": False})
        self._access_token = access_token
        self._expiry = self._claims["exp"]

        return access_token

    def _parse_token(self, token_response: Dict[str, str]) -> str:
        """Parse the token from the token response

        Args:
            The token response from the IDP

        Returns:
            str: The token
        """
        try:
            access_token = token_response["access_token"]
        except KeyError as e:
            raise OdpAuthError("No access token in token response") from e

        if not self.validate_token:
            return self._parse_novalidate(access_token)
        return self._parse_and_validate(access_token)

    def _get_jwk(self, kid: str) -> Dict:
        jwks = self._get_jwks()
        for jwk in jwks["keys"]:
            if jwk.get("kid") == kid:
                return jwk

        raise OdpAuthError("Invalid token: Invalid KID")

    def _get_jwks(self) -> Dict:
        if self._jwks:
            return self._jwks

        jwks_uri = self.get_jwks_uri()

        res = requests.get(jwks_uri)
        res.raise_for_status()

        self._jwks = res.json()
        return self._jwks

    @staticmethod
    def _base64_decode(data: str, altchars: bytes = b"+/") -> bytes:
        d = re.sub(rb"[^a-zA-Z0-9%s]+" % altchars, b"", data.encode("utf-8"))

        if len(d) % 4 == 1:
            d = d[:-1]
        else:
            # Add padding
            d += (-len(d) % 4) * b"="
        decoded = base64.b64decode(d, altchars=altchars)
        return decoded

    @staticmethod
    def _decode_jwk_value(val: str) -> int:
        decoded = base64.urlsafe_b64decode(val.encode("utf-8") + b"==")
        return int.from_bytes(decoded, "big")


class AzureTokenProvider(JwtTokenProvider):
    """Token provider for OAuth2 tokens"""

    client_id: SecretStr
    """IDP client ID"""

    client_secret: SecretStr
    """IDP client secret"""

    authority: str = "https://oceandataplatform.b2clogin.com/755f6e58-74f0-4a07-a599-f7479b9669ab/v2.0/"
    """IDP token Authority"""

    scope: List[str] = ["https://oceandataplatform.onmicrosoft.com/odcat/.default"]
    """IDP token scope"""

    tenant_id: str = "755f6e58-74f0-4a07-a599-f7479b9669ab"
    """Azure AD tenant id"""

    token_uri: str = "https://oceandataplatform.b2clogin.com/oceandataplatform.onmicrosoft.com/b2c_1a_signup_signin_custom/oauth2/v2.0/token"  # noqa: E501
    """Token endpoint. Will default to 'https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token'"""

    jwks_uri: str = "https://oceandataplatform.b2clogin.com/oceandataplatform.onmicrosoft.com/b2c_1a_signup_signin_custom/discovery/v2.0/keys"  # noqa: E501
    """JWKS endpoint."""

    def __init__(self, **data):
        super().__init__(**data)
        self.user_agent = self.user_agent + " (Azure)"

    def get_jwks_uri(self) -> str:
        return self.jwks_uri

    def authenticate(self) -> Dict[str, str]:
        res = requests.post(
            self.token_uri,
            data={
                "grant_type": "client_credentials",
                "client_id": self.client_id.get_secret_value(),
                "client_secret": self.client_secret.get_secret_value(),
                "audience": self.audience,
                "scope": " ".join(self.scope),
            },
        )

        res.raise_for_status()
        return res.json()


class InteractiveTokenProvider(JwtTokenProvider):
    """Token provider for interactive token input

    This token provider is intended to be used for interactive sessions, where the user is prompted for
    login credentials by a web browser. The token is then retrieved from the browser and used for authentication.
    """

    client_id: SecretStr
    """IDP client ID"""

    authority: str = (
        "https://oceandataplatform.b2clogin.com/oceandataplatform.onmicrosoft.com/b2c_1a_signup_signin_custom"
    )
    """IDP token Authority"""

    scope: List[str] = ["https://oceandataplatform.onmicrosoft.com/odcat/API_ACCESS"]
    """IDP token scope"""

    jwks_uri: str = "https://oceandataplatform.b2clogin.com/oceandataplatform.onmicrosoft.com/b2c_1a_signup_signin_custom/discovery/v2.0/keys"  # noqa: E501
    """JWKS endpoint."""

    token_persistence_file: Path = Path(".token_cache.bin")
    """File to save the token cache to"""

    token_persistence_plaintext_fallback: bool = True
    """Whether to fallback to plaintext if encryption is unavailable"""

    _app: msal.PublicClientApplication = PrivateAttr(None)

    def __init__(self, **data):
        super().__init__(**data)

        persistence = self._build_persistence(self.token_persistence_file, self.token_persistence_plaintext_fallback)
        cache = msal_extensions.PersistedTokenCache(persistence)

        self._app = msal.PublicClientApplication(
            client_id=self.client_id.get_secret_value(),
            authority=self.authority,
            token_cache=cache,
        )
        self.user_agent = self.user_agent + " (Interactive)"

    def get_jwks_uri(self) -> str:
        return self.jwks_uri

    def authenticate(self) -> Dict[str, str]:
        accounts = self._app.get_accounts()
        if accounts and len(accounts) == 1:
            account = accounts[0]

            LOG.info(
                "Found account %s. Attempting to acquire token silently",
                account["username"],
            )
            token = self._app.acquire_token_silent(
                scopes=self.scope,
                account=account,
                force_refresh=True,
            )
        else:
            token = None

        if token is None:
            LOG.info("No suitable token found. Acquiring token interactively")
            res = self._app.acquire_token_interactive(
                scopes=self.scope,
            )
            if "error" in res:
                raise OdpAuthError("{}: {}".format(res["error"], res["error_description"]))

            # Token acquired successfully, rerun the method to get the token
            return self.authenticate()

        return token

    @staticmethod
    def _build_persistence(location: Path, fallback_to_plaintext: bool = False):
        if sys.platform.startswith("win"):
            return msal_extensions.FilePersistenceWithDataProtection(location)
        elif sys.platform.startswith("darwin"):
            return msal_extensions.KeychainPersistence(location, "odcat", "odcat-user")
        elif sys.platform.startswith("linux"):
            try:
                return msal_extensions.LibsecretPersistence(
                    str(location),
                    schema_name="odcat-sdk",
                    attributes={"app": "odcat", "component": "odcat-cli"},
                )
            except Exception:
                if not fallback_to_plaintext:
                    raise
                LOG.warning("Encryption unavailable. Opting in to plaintext")
        return msal_extensions.FilePersistence(str(location))


def get_default_token_provider(fallback: bool = True) -> TokenProvider:
    """Automatically select a token provider based on the environment

    Args:
        fallback: Whether to fallback to an interactive token provider if no other token provider is available

    Returns:
        The token provider

    Raises:
        OdpAuthError: If no token provider is available and fallback is False
    """
    if hardcoded_access_token := os.getenv("ODP_ACCESS_TOKEN"):
        return HardcodedTokenProvider(hardcoded_access_token)

    if os.getenv("JUPYTERHUB_API_TOKEN"):
        return OdpWorkspaceTokenProvider()

    client_id = os.getenv("ODP_CLIENT_ID", "f96fc4a5-195b-43cc-adb2-10506a82bb82")
    client_secret = os.getenv("ODP_CLIENT_SECRET")

    if client_secret:
        return AzureTokenProvider(client_id=SecretStr(client_id), client_secret=SecretStr(client_secret))

    if fallback:
        return InteractiveTokenProvider(client_id=SecretStr(client_id))

    raise OdpAuthError("Unable to find a suitable token provider")



from odp.util.check_version import check_library_version

# FIXME: This is only for compatibility with pydantic and should be removed
if check_library_version("pydantic") == "1.10.20":
    from pydantic import BaseModel, Field, PrivateAttr, SecretStr
else:
    from .cheapdantic import BaseModel, Field, PrivateAttr, SecretStr



import importlib.metadata
import importlib.util


def check_library_version(library_name: str):
    # Check if the library is present
    spec = importlib.util.find_spec(library_name)
    if spec is not None:
        # If present, get the version
        try:
            version = importlib.metadata.version(library_name)
            return version
        except importlib.metadata.PackageNotFoundError:
            pass



from typing import Any, Dict

import jsonschema

from ..resource import ResourceDto, ResourceSpec
from ..resource_registry import DEFAULT_RESOURCE_TYPE_REGISTRY
from ..validators import validate_json_schema


class ObservableClassSpec(ResourceSpec):
    observable_schema: Dict
    """JSON Schema for the observable class"""

    def __init__(self, **kwargs):
        kwargs["observable_schema"] = validate_json_schema(kwargs.get("observable_schema"))
        super().__init__(**kwargs)

    def validate_dict(self, d: Dict[str, Any]):
        try:
            jsonschema.validate(instance=d, schema=self.observable_schema)
        except jsonschema.ValidationError as e:
            raise ValueError(f"Observable details do not match schema: {e.message}") from e


class ObservableClassDto(ResourceDto):
    _kind: str = "catalog.hubocean.io/observableClass"
    _version: str = "v1alpha1"

    spec: ObservableClassSpec


DEFAULT_RESOURCE_TYPE_REGISTRY.add(ObservableClassDto)



import odp.tabular_v2.client as tab
from odp.new_client import Client


class Dataset:
    """
    Represents a dataset in the Ocean Data Platform.
    This class provides access to the dataset's storage (files, table)
    """

    def __init__(self, cli: Client, id: str):
        self.cli = cli
        self.id = id
        self.table = tab.Table(cli, id)
        self.files = self.table.raw
        self.name = ""  # do integration with catalog v2 when ready

        # shortcuts for common operations
        self.select = self.table.select

    def __str__(self):
        if self.name:
            return f"Dataset({self.id} - {self.name})"
        else:
            return f"Dataset({self.id})"

    def __enter__(self):
        return self.table.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.table.__exit__(exc_type, exc_val, exc_tb)

    @staticmethod
    def _from_catalog_v2(cli: Client, data: dict):
        ds = Dataset(cli, data["id"])
        ds.name = data["name"]
        return ds
